# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click

from .commands import decrypt, download, token


@click.group()
def cli():
    pass


cli.add_command(token.token)
cli.add_command(download.download)
cli.add_command(decrypt.decrypt)

if __name__ == "__main__":
    cli()



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import sys

import click
import gpsoauth

from .. import gpsoauth_helper
from ..constants import MASTER_TOKEN_SUFFIX, TOKEN_SUFFIX, TOKENS_FOLDER
from ..utils import generate_android_uid


@click.command(name="token")
@click.argument("email")
@click.option("--token-file", help="Token file", default=None)
@click.option("--master-token", help="Master token file", default=None)
@click.option("--android-id", help="Android ID", default=None)
def token(token_file: str, master_token: str, android_id: str, email: str):
    # Generate a random android_id if not provided
    if android_id is None:
        android_id = generate_android_uid()

    # Use the default token file if not provided
    if token_file is None:
        token_file = email.replace("@", "_").replace(".", "_") + TOKEN_SUFFIX

        # Create tokens folder if it doesn't exist
        tokens_folder = pathlib.Path.cwd() / TOKENS_FOLDER
        tokens_folder.mkdir(parents=True, exist_ok=True)
        token_filepath = tokens_folder / token_file
    else:
        token_filepath = pathlib.Path(token_file)

    # Use the default master token file if not provided
    if master_token is None:
        master_token_file = email.replace("@", "_").replace(".", "_") + MASTER_TOKEN_SUFFIX

        # Create tokens folder if it doesn't exist
        tokens_folder = pathlib.Path.cwd() / TOKENS_FOLDER
        tokens_folder.mkdir(parents=True, exist_ok=True)
        master_token_filepath = tokens_folder / master_token_file
    else:
        master_token_filepath = pathlib.Path(master_token)

    # Ask for the oauth_token cookie
    print("Please visit https://accounts.google.com/EmbeddedSetup, login and copy the oauth_token cookie.")
    oauth_token = input('Enter "oauth_token" code: ')

    # Exchange the token for a master token
    master_response = gpsoauth.exchange_token(email, oauth_token, android_id)
    if "Error" in master_response:
        print(master_response["Error"], file=sys.stderr)
        sys.exit(1)

    master_token = master_response["Token"]
    with open(master_token_filepath, "w") as f:
        f.write(master_token)
        print(f"Master Token saved to `{master_token_filepath}`")

    try:
        auth_token = gpsoauth_helper.get_auth_token(master_token, android_id)
    except gpsoauth_helper.AuthException as e:
        print(e, file=sys.stderr)
        print(e.get_extra(), file=sys.stderr)
        sys.exit(1)

    # Save the token to a file
    with open(token_filepath, "w") as f:
        f.write(auth_token)
        print(f"Token saved to `{token_filepath}`")

    print(f"You can now run `wabdd download --master-token {master_token_filepath}`")



# Copyright 2025 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest.mock import patch

from wabdd.utils import crop_string, generate_android_uid, sizeof_fmt


def test_bytes():
    assert sizeof_fmt(0) == "0.0B"
    assert sizeof_fmt(1) == "1.0B"
    assert sizeof_fmt(500) == "500.0B"
    assert sizeof_fmt(1023) == "1023.0B"


def test_kilobytes():
    assert sizeof_fmt(1024) == "1.0KiB"


def test_megabytes():
    assert sizeof_fmt(1024 * 1024) == "1.0MiB"


def test_gigabytes():
    assert sizeof_fmt(1024 * 1024 * 1024) == "1.0GiB"


def test_terabytes():
    assert sizeof_fmt(1024 * 1024 * 1024 * 1024) == "1.0TiB"


def test_petabytes():
    assert sizeof_fmt(1024 * 1024 * 1024 * 1024 * 1024) == "1.0PiB"


def test_exabytes():
    assert sizeof_fmt(1024 * 1024 * 1024 * 1024 * 1024 * 1024) == "1.0EiB"


def test_zettabytes():
    assert sizeof_fmt(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024) == "1.0ZiB"


def test_yottabytes():
    assert sizeof_fmt(1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024 * 1024) == "1.0YiB"


def test_crop_string():
    assert crop_string("Hello, World!", 5) == "…rld!"


def test_crop_string_no_crop():
    assert crop_string("Hello, World!", 20) == "Hello, World!"


def test_crop_string_empty():
    assert crop_string("", 5) == ""


def test_crop_string_ellipsis():
    assert crop_string("Hello, World!", 5, "...", False) == "...orld!"


def test_crop_string_no_ellipsis():
    assert crop_string("Hello, World!", 5, "") == "orld!"


def test_crop_string_long_ellipsis():
    assert crop_string("Hello, World!", 5, "....", False) == "....orld!"


def test_generate_android_uid():
    with patch("random.getrandbits", return_value=0x123456789ABCDEF0):
        uid = generate_android_uid()
        assert uid == "123456789abcdef0"



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from urllib.parse import quote

import requests

from . import gpsoauth_helper, utils
from .constants import USER_AGENT


class WaBackup:
    def __init__(self, auth_token, master_token):
        self.auth_token = auth_token
        self.master_token = master_token

        # Check if at least one token is provided
        if not (auth_token or master_token):
            raise ValueError("At least one token must be provided")

    def _get(self, path, params=None, **kwargs):
        path = quote(path)
        r = requests.get(
            f"https://backup.googleapis.com/v1/{path}",
            headers={
                "Authorization": f"Bearer {self.auth_token}",
                "User-Agent": USER_AGENT,
            },
            params=params,
            **kwargs,
        )

        # If the request is unauthorized, try to refresh the auth token
        if r.status_code == 401:
            if self.master_token is None:
                raise ValueError(
                    "Auth token is expired and no master token is provided. "
                    "You need to provide a master token to refresh the auth token."
                )

            try:
                android_id = utils.generate_android_uid()
                self.auth_token = gpsoauth_helper.get_auth_token(self.master_token, android_id)
            except gpsoauth_helper.AuthException:
                raise ValueError("Something went wrong while refreshing the auth token")

            r = requests.get(
                f"https://backup.googleapis.com/v1/{path}",
                headers={
                    "Authorization": f"Bearer {self.auth_token}",
                    "User-Agent": USER_AGENT,
                },
                params=params,
                **kwargs,
            )

        return r

    def _get_page(self, path, page_token=None):
        return self._get(
            path,
            None if page_token is None else {"pageToken": page_token},
        ).json()

    def download(self, path):
        return self._get(
            path,
            params={"alt": "media"},
            stream=True,
        )

    def _list_path(self, path):
        last_component = path.split("/")[-1]
        page_token = None
        while True:
            page = self._get_page(path, page_token)

            # Early exit if no key is found (e.g. no backups)
            if last_component not in page:
                break

            # Yield each item in the page
            for item in page[last_component]:
                yield item

            # If there is no nextPageToken, we are done
            if "nextPageToken" not in page:
                break

            page_token = page["nextPageToken"]

    def get_backups(self):
        return self._list_path("clients/wa/backups")

    def backup_files(self, backup):
        return self._list_path("{}/files".format(backup["name"]))



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import fnmatch
import json
import pathlib
import sys
import traceback
from datetime import datetime
from queue import Queue
from threading import Event, Thread
from time import time
from typing import Optional, Tuple

import click
import inquirer
import requests
from rich.console import Group
from rich.live import Live
from rich.panel import Panel
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    ProgressColumn,
    SpinnerColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.text import Text
from wa_crypt_tools.lib.key.key15 import Key15
from wa_crypt_tools.lib.utils import mcrypt1_metadata_decrypt

from ..constants import BACKUP_FOLDER
from ..utils import crop_string, get_md5_hash_from_file, sizeof_fmt
from ..wabackup import WaBackup


class ItemSpeedColumn(ProgressColumn):
    """Custom column to show speed in items per second."""

    def __init__(self):
        super().__init__()
        self.previous_completed = {}
        self.previous_time = {}

    def render(self, task):
        # Get the current time and completed items for the task
        current_time = time()
        completed = task.completed

        # Calculate time elapsed since the last update
        last_completed = self.previous_completed.get(task.id, 0)
        last_time = self.previous_time.get(task.id, current_time)

        elapsed_time = current_time - last_time if last_time != current_time else 1e-10  # avoid division by zero

        # Calculate speed in items per second
        speed = (completed - last_completed) / elapsed_time if elapsed_time > 0 else 0

        # Update the previous values for next calculation
        self.previous_completed[task.id] = completed
        self.previous_time[task.id] = current_time

        return Text(f"{speed:.0f} it/s", style="progress.percentage")


_sentinel = object()
_stop_event = Event()


class DownloaderWorker(Thread):
    def __init__(
        self,
        queue: Queue,
        output: pathlib.Path,
        progress: Progress,
        overall_progress: Tuple[Progress, TaskID],
        client: WaBackup,
        exclude_pattern: Optional[Tuple[str]] = None,
        decryption_key: Optional[Key15] = None,
    ):
        super().__init__()
        self.queue = queue
        self.output = output
        self.progress = progress
        self.overall_progress = overall_progress[0]
        self.overall_task = overall_progress[1]
        self.client = client
        self.exclude_pattern = exclude_pattern
        self.decryption_key = decryption_key

        self.is_running = True

    def run(self):
        while self.is_running:
            self.task_id = None

            if _stop_event.is_set():
                self.is_running = False
                return

            try:
                # Get a file from the queue
                file = self.queue.get()

                # Check if we're done
                if file is _sentinel:
                    self.is_running = False
                    return

                file_path = file["path"]
                file_name = file_path.split("/")[-1]
                file_hash = file["hash"]
                file_size = file["size"]
                metadata = file["metadata"]

                # Strip first 5 components of the path
                # (e.g. clients/wa/backups/00000/files)
                stripped_filepath = pathlib.Path(*file_path.split("/")[5:])

                # Decrypt metadata if available
                real_path = str(stripped_filepath)
                if self.decryption_key and metadata:
                    decrypted_metadata = mcrypt1_metadata_decrypt(key=self.decryption_key, encoded=metadata)
                    real_path = str(decrypted_metadata.get("name", real_path))

                # Check if file should be excluded
                if self.exclude_pattern:
                    if any(fnmatch.fnmatch(real_path, pattern) for pattern in self.exclude_pattern):
                        self.progress.console.print(f"Skipping {real_path} (excluded)")
                        continue

                # Check if file is already downloaded
                local_file = self.output / stripped_filepath
                if local_file.exists():
                    if get_md5_hash_from_file(local_file) == file_hash:
                        # self.progress.console.print(f"Skipping {file_name}")
                        continue

                # Create task
                self.task_id = self.progress.add_task(
                    "",
                    start=False,
                    filename=crop_string(file_name, 20),
                    total=file_size,
                )

                # Download the file
                with self.client.download(file_path) as r:
                    if r.status_code == 401:
                        self.progress.console.print(
                            "Token expired, stopping download",
                        )
                        _stop_event.set()
                        raise Exception("Unauthorized (401) - Stopping all workers")

                    if r.status_code == 404:
                        self.progress.console.print(
                            f"Error: File {file_name} not found",
                        )
                        continue

                    if r.status_code == 400:
                        self.progress.console.print(
                            f"Error: Bad request for file {file_name}",
                        )
                        continue

                    r.raise_for_status()
                    with open(local_file, "wb") as f:
                        self.progress.start_task(self.task_id)
                        for chunk in r.iter_content(chunk_size=8192):
                            if _stop_event.is_set():  # Exit early if stop event is set
                                return

                            self.progress.update(self.task_id, advance=len(chunk))
                            f.write(chunk)

                # Write metadata if available
                if metadata:
                    with open(str(local_file) + "-metadata", "w") as f:
                        f.write(metadata)

            except Exception as e:
                self.progress.console.print(f"Error: {e}")
                self.progress.console.print(traceback.format_exc())
                self.is_running = False
            finally:
                # Stop the download progress task
                if hasattr(self, "task_id") and self.task_id is not None:
                    self.progress.remove_task(self.task_id)

                # Update overall progress
                if self.is_running:
                    self.overall_progress.update(
                        self.overall_task,
                        advance=1,
                    )

                # Indicate task completion
                self.queue.task_done()


@click.command(name="download")
@click.option("--token-file", help="Token file", default=None)
@click.option("--master-token", help="Master token file", default=None)
@click.option("--output", help="Output directory")
@click.option("--threads", help="Number of threads to use", default=8)
@click.option(
    "--save-files-list",
    help="Save list of files to files.json",
    is_flag=True,
    default=False,
)
@click.option(
    "--save-backup-metadata",
    help="Save backup metadata to metadata.json",
    is_flag=True,
    default=True,
)
@click.option(
    "--exclude",
    "exclude_pattern",
    help="Exclude files matching the given pattern",
    multiple=True,
)
@click.option("--decryption-key-file", help="Key file to use for decryption", default=None)
def download(
    token_file: str,
    master_token: str,
    output: str,
    threads: int,
    save_files_list: bool,
    save_backup_metadata: bool,
    exclude_pattern: Tuple[str],
    decryption_key_file: str,
):
    # Check for token file or master token
    if not token_file and not master_token:
        print("Please provide a token file or a master token", file=sys.stderr)
        sys.exit(1)

    # Check for token
    if token_file:
        if not pathlib.Path(token_file).exists():
            print(f"Token file {token_file} not found", file=sys.stderr)
            print("Run `wabdd token` to generate a token", file=sys.stderr)
            sys.exit(1)

        # Load token
        with open(token_file) as f:
            token = f.read().strip()
    else:
        token = None

    # Check for master token
    if master_token:
        if not pathlib.Path(master_token).exists():
            print(f"Master token file {master_token} not found", file=sys.stderr)
            print("Run `wabdd token` to generate a master token", file=sys.stderr)
            sys.exit(1)

        # Load master token
        with open(master_token) as f:
            master_token = f.read().strip()

    # Load decryption key
    decryption_key = None
    if decryption_key_file:
        with open(decryption_key_file, "r") as f:
            key_bytes = bytes.fromhex(f.read())
            if len(key_bytes) != 32:
                raise ValueError("Key must be 32 bytes long")

            decryption_key = Key15(key_bytes)

    # Initialize client
    wa = WaBackup(token, master_token)

    # Get backups
    try:
        _backups = list(wa.get_backups())
    except requests.exceptions.HTTPError as e:
        print(f"Error: {e}")
        print("Run `wabdd token` to generate a new token", file=sys.stderr)
        sys.exit(1)

    # Filter out backups that are still uploading
    backups = []
    for backup in _backups:
        if "activeTransactionId" in backup:
            print(f"Backup {backup['name']} is still uploading, skipping")
            print(json.dumps(backup, indent=2))
            continue

        backups.append(backup)

    if len(backups) == 0:
        print("No backups found")
        return

    if len(backups) > 1:
        for i, backup in enumerate(backups):
            print(f"Backup {i + 1}: {backup['name']} ({backup['updateTime']})")

        print("Multiple backups found, please specify which one to download")

        questions = [
            inquirer.List(
                "backup",
                message="Which backup do you want to download?",
                choices=[
                    (
                        f"Backup {i + 1}: {backup['name'].split('/')[-1]} ({backup['updateTime']})",
                        backup,
                    )
                    for i, backup in enumerate(backups)
                ],
            ),
        ]
        answers = inquirer.prompt(questions)

        # Exit if no backup is selected
        if not answers:
            print("No backup selected")
            return

        backup = answers["backup"]
    else:
        backup = backups[0]

    # Get backup metadata
    backup_metadata = json.loads(
        backup.get(
            "metadata",
            json.dumps(
                {
                    "encryptedBackupEnabled": False,
                    "backupSize": 0,
                    "chatdbSize": 0,
                }
            ),
        )
    )

    # Build exclude pattern
    is_encrypted_backup = backup_metadata.get("encryptedBackupEnabled", False)
    if len(exclude_pattern) > 0 and is_encrypted_backup and decryption_key is None:
        print(
            "WARNING: --exclude patterns will not work with encrypted backups without a decryption key."
            "Please provide the decryption key file with --decryption-key-file",
            file=sys.stderr,
        )

    # Print backup information
    print(f"Backup: {backup['name'].split('/')[-1]} ({backup['updateTime']})")
    print(" - Total size:", sizeof_fmt(backup_metadata["backupSize"]))
    print(" - Chat size:", sizeof_fmt(backup_metadata["chatdbSize"]))

    # Set output directory
    if not output:
        timestamp = datetime.strptime(backup["updateTime"], "%Y-%m-%dT%H:%M:%S.%fZ")
        output_dir = (
            pathlib.Path.cwd() / BACKUP_FOLDER / f"{backup['name'].split('/')[-1]}_{timestamp.strftime('%Y%m%d')}"
        )
    else:
        output_dir = pathlib.Path(output)

    # Check if the output dir is a file
    if output_dir.exists() and not output_dir.is_dir():
        print(f"Error: {output_dir} is not a directory", file=sys.stderr)
        sys.exit(1)

    # Create output directory
    output_dir.mkdir(parents=True, exist_ok=True)

    # Write metadata
    if save_backup_metadata:
        with open(output_dir / "metadata.json", "w") as f:
            json.dump(backup_metadata, f, separators=(",", ":"))

    print("Downloading to", output_dir)

    # Download files
    file_retrieval_progress = Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        DownloadColumn(binary_units=True),
        TimeElapsedColumn(),
    )

    download_overall_progress = Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        TextColumn("[yellow]{task.completed}/{task.total}"),
        ItemSpeedColumn(),
        TaskProgressColumn(),
        TimeElapsedColumn(),
    )

    download_progress = Progress(
        TextColumn("[bold yellow]Downloading {task.fields[filename]}"),
        BarColumn(bar_width=None),
        DownloadColumn(binary_units=True),
        TransferSpeedColumn(),
        TimeRemainingColumn(),
    )

    overall_progress = Progress(TimeElapsedColumn(), BarColumn(bar_width=None), TextColumn("{task.description}"))

    group = Group(
        Panel(
            Group(
                file_retrieval_progress,
                download_overall_progress,
                download_progress,
            )
        ),
        overall_progress,
    )

    steps = [
        "Retrieving list of files...",
        "Preparing folders...",
        "Downloading files...",
    ]
    current_step = 0

    overall_task_id = overall_progress.add_task("", total=len(steps))

    with Live(group):
        # Retrieve the list of files
        overall_progress.update(
            overall_task_id,
            description=f"{steps[current_step]} ({current_step + 1}/{len(steps)})",
        )
        task_id = file_retrieval_progress.add_task(
            "Retrieving list of files...", total=int(backup.get("sizeBytes", "0"))
        )
        files = []
        debug_files = []
        for file in wa.backup_files(backup):
            if save_files_list:
                debug_files.append(file)

            files.append(
                {
                    "path": file.get("name"),
                    "hash": base64.b64decode(file.get("md5Hash")),
                    "size": int(file.get("sizeBytes")),
                    "metadata": file.get("metadata"),
                }
            )
            file_retrieval_progress.update(
                task_id,
                advance=int(file.get("sizeBytes")),
            )

        file_retrieval_progress.stop_task(task_id)
        file_retrieval_progress.update(task_id, description=f"[bold green]All {len(files)} files retrieved!")

        # Debug
        if save_files_list:
            with open(output_dir / "files.json", "w") as f:
                json.dump(debug_files, f, separators=(",", ":"))

        current_step += 1
        overall_progress.update(overall_task_id, advance=1)

        # Prepare directories
        overall_progress.update(
            overall_task_id,
            description=f"{steps[current_step]} ({current_step + 1}/{len(steps)})",
        )
        already_created = set()
        for file in files:
            target: pathlib.Path = output_dir / pathlib.Path(*file["path"].split("/")[5:]).parent

            # Skip if already created
            if target in already_created:
                continue

            # Check if target is a directory
            if target.exists() and not target.is_dir():
                print(f"Error: {target} is not a directory", file=sys.stderr)
                sys.exit(1)

            target.mkdir(parents=True, exist_ok=True)
            already_created.add(target)

        current_step += 1
        overall_progress.update(overall_task_id, advance=1)

        # Download files
        overall_progress.update(
            overall_task_id,
            description=f"{steps[current_step]} ({current_step + 1}/{len(steps)})",
        )

        overall_download_task_id = download_overall_progress.add_task("Downloading files...", total=len(files))

        queue = Queue()
        for file in files:
            queue.put(file)

        workers = []
        for _ in range(threads):
            # Add a sentinel to the queue to indicate the end of the queue
            queue.put(_sentinel)

            worker = DownloaderWorker(
                queue,
                output_dir,
                download_progress,
                (download_overall_progress, overall_download_task_id),
                wa,
                exclude_pattern,
                decryption_key,
            )
            worker.start()
            workers.append(worker)

        # TODO: won't work if uncommented
        # Wait for all files to download
        # queue.join()

        # Wait for all workers to finish
        try:
            for worker in workers:
                worker.join()
        except KeyboardInterrupt:
            print("Keyboard Interrupt received! Stopping workers...")
            _stop_event.set()

            # TODO: not sure if this is necessary
            # Wait for remaining workers to finish
            for worker in workers:
                worker.join()

        download_overall_progress.stop_task(overall_download_task_id)
        download_overall_progress.update(
            overall_download_task_id,
            description=f"[bold green]All {len(files)} files downloaded!",
        )

        # Complete progress
        overall_progress.update(
            overall_task_id,
            advance=1,
            description=f"FINISHED! ({current_step + 1}/{len(steps)})",
        )

    print(f"You can now run `wabdd decrypt --key-file YOUR_KEY_FILE dump {output_dir}`")



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import pathlib
import random
from typing import Union


# https://stackoverflow.com/questions/1094841/
def sizeof_fmt(num: float, suffix="B"):
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f%s%s" % (num, "Yi", suffix)


def generate_android_uid():
    return f"{random.getrandbits(64):016x}"


def get_md5_hash_from_file(file: Union[str, pathlib.Path]):
    with open(file, "rb") as f:
        file_hash = hashlib.md5()
        while chunk := f.read(8192):
            file_hash.update(chunk)

    return file_hash.digest()


def crop_string(s: str, N: int, ellipsis="…", include_ellipsis=True):
    ellipsis_length = len(ellipsis) if include_ellipsis else 0

    # Check if the string needs to be cropped
    if len(s) > N - ellipsis_length:
        return ellipsis + s[(-(N - ellipsis_length)) :]
    else:
        return s



# Copyright 2024 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pathlib
import sys
import traceback
import zlib
from datetime import datetime
from queue import Queue
from threading import Event, Thread
from typing import Tuple

import click
from Cryptodome.Cipher import AES
from rich.console import Group
from rich.live import Live
from rich.progress import (
    BarColumn,
    Progress,
    SpinnerColumn,
    TaskID,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)
from wa_crypt_tools.lib.db.dbfactory import Database15, DatabaseFactory
from wa_crypt_tools.lib.key.key15 import Key15
from wa_crypt_tools.lib.utils import encryptionloop, mcrypt1_metadata_decrypt

_sentinel = object()
_stop_event = Event()


class DecryptionWorker(Thread):
    def __init__(
        self,
        queue: Queue,
        output: pathlib.Path,
        overall_progress: Tuple[Progress, TaskID],
        key: Key15,
    ):
        super().__init__()
        self.queue = queue
        self.output = output
        self.overall_progress = overall_progress[0]
        self.overall_task = overall_progress[1]
        self.key = key

        self.is_running = True

    def run(self):
        while self.is_running:
            self.task_id = None

            if _stop_event.is_set():
                self.is_running = False
                return

            try:
                # Get a file from the queue
                task = self.queue.get()

                # Check if we're done
                if task is _sentinel:
                    self.is_running = False
                    return

                folder, file = task
                self.ERROR_FOLDER = folder
                self.ERROR_FILE = file

                # Decrypt file
                if file.suffix == ".mcrypt1":
                    output_file, decrypted_data, file_timestamp = decrypt_mcrypt1_file(folder, file, self.key)
                elif file.suffix == ".crypt15":
                    output_file, decrypted_data, file_timestamp = decrypt_crypt15_file(folder, file, self.key)
                else:
                    # Copy unsupported files
                    self.overall_progress.console.print(f"Unsupported file: {file}")
                    output_file = file
                    decrypted_data = output_file.read_bytes()

                # Write decrypted data to output file
                (self.output / output_file).parent.mkdir(parents=True, exist_ok=True)
                with open(self.output / output_file, "wb") as f:
                    f.write(decrypted_data)

                # Update file timestamp
                if file_timestamp:
                    timestamp = file_timestamp.timestamp()
                    os.utime((self.output / output_file), (timestamp, timestamp))

            except Exception as e:
                self.overall_progress.console.print(f"Error in {self.ERROR_FOLDER} {self.ERROR_FILE}")
                self.overall_progress.console.print(f"Error: {e}")
                self.overall_progress.console.print(traceback.format_exc())
                self.is_running = False
            finally:
                # Update overall progress
                if self.is_running:
                    self.overall_progress.update(
                        self.overall_task,
                        advance=1,
                    )

                # Indicate task completion
                self.queue.task_done()


def decrypt_metadata(metadata_file: pathlib.Path, key: Key15):
    with open(metadata_file) as f:
        return mcrypt1_metadata_decrypt(key=key, encoded=f.read())


def decrypt_mcrypt1_file(
    dump_folder: pathlib.Path, encrypted_file: pathlib.Path, key: Key15
) -> Tuple[pathlib.Path, bytes, datetime]:
    # Get filename without `.mcrypt1` extension and convert to bytes
    decryption_hash = bytes.fromhex(encrypted_file.with_suffix("").name)
    decryption_data = encryptionloop(
        first_iteration_data=key.get_root(),
        message=decryption_hash,
        output_bytes=48,
    )

    # Prepare AES
    aes_key = decryption_data[:32]
    aes_iv = decryption_data[32:48]
    cipher = AES.new(aes_key, AES.MODE_GCM, aes_iv)

    # Get metadata
    output_file = encrypted_file.with_suffix("")
    metadata_file = encrypted_file.with_suffix(".mcrypt1-metadata")
    if metadata_file.is_file():
        metadata = decrypt_metadata(metadata_file, key)
        output_file = pathlib.Path(metadata["name"])

    with open(encrypted_file, "rb") as f:
        return (
            output_file,
            cipher.decrypt(f.read()),
            datetime.fromisoformat(metadata["updateTime"]),
        )


def decrypt_crypt15_file(
    dump_folder: pathlib.Path, encrypted_file: pathlib.Path, key: Key15
) -> Tuple[pathlib.Path, bytes, None]:
    output_file = encrypted_file.relative_to(dump_folder)

    # Remove .crypt15 extension
    output_file = output_file.with_suffix("")

    # Open .crypt15 file
    with open(encrypted_file, "rb") as f:
        db = DatabaseFactory.from_file(f)

        # Check if db is Database15
        if not isinstance(db, Database15):
            raise ValueError("Unsupported database format")

        decrypted_data = db.decrypt(key, f.read())

    # Try to decompress the data
    try:
        z_obj = zlib.decompressobj()
        decrypted_data = z_obj.decompress(decrypted_data)

        if not z_obj.eof:
            print(
                f"WARNING: There is more data to decompress. {output_file}",
                file=sys.stderr,
            )
    except zlib.error:
        decrypted_data = decrypted_data

    return output_file, decrypted_data, None


@click.group()
@click.option("--key", help="Key to use for decryption", default=None)
@click.option("--key-file", help="Key file to use for decryption", default=None)
@click.pass_context
def decrypt(ctx, key, key_file):
    if key is None and key_file is None:
        print("Please provide either a --key or a --key-file", file=sys.stderr)
        sys.exit(1)

    if key is not None and key_file is not None:
        print("Please provide either a --key or a --key-file, not both", file=sys.stderr)
        sys.exit(1)

    if key_file is not None:
        try:
            with open(key_file, "r") as f:
                key = f.read().strip()
        except Exception as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    try:
        key_bytes = bytes.fromhex(key)

        if len(key_bytes) != 32:
            raise ValueError("Key must be 32 bytes long")

        ctx.obj = Key15(keyarray=key_bytes)
    except ValueError as e:
        print(f"Error: {e}", file=sys.stderr)
        sys.exit(1)


@decrypt.command(name="dump")
@click.argument("folder", type=click.Path(exists=True))
@click.option("--output", help="Output directory", default=None)
@click.option("--threads", help="Number of threads to use", default=2)
@click.pass_obj
def cmd_decrypt_dump(obj, folder, output, threads):
    key = obj
    folder = pathlib.Path(folder)

    # Set default output directory
    if not output:
        output = f"{folder}-decrypted"

    # Create output directory
    output = pathlib.Path(output)
    output.mkdir(parents=True, exist_ok=True)

    # Check for correct dump structure
    paths_to_check = [
        "Databases",
        "Backups",
        # "Media", # Not necessary if user has no media
        # "metadata.json",
    ]
    for path in paths_to_check:
        if not (folder / path).exists():
            print(f"Error: {folder / path} not found", file=sys.stderr)
            sys.exit(1)

    # 3.96s user 2.98s system 95% cpu 7.236 total baseline
    # 3.91s user 3.01s system 95% cpu 7.251 total n=1
    # 3.86s user 1.72s system 157% cpu 3.554 total n=2
    # 4.59s user 2.76s system 209% cpu 3.513 total n=4
    # Decrypt files
    queue = Queue()

    download_overall_progress = Progress(
        SpinnerColumn(),
        TextColumn("[bold blue]{task.description}", justify="right"),
        BarColumn(bar_width=None),
        TextColumn("[yellow]{task.completed}/{task.total}"),
        TaskProgressColumn(),
        TimeElapsedColumn(),
    )

    group = Group(download_overall_progress)

    with Live(group):
        # Prepare queue
        queue = Queue()
        wabdd_files = ["metadata.json", "files.json"]
        supported_extensions = [".crypt15", ".mcrypt1", ".mcrypt1-metadata"]
        ignored_extensions = [".mcrypt1-metadata"]
        for file in folder.glob("**/*"):
            # Skip folder
            if file.is_dir():
                continue

            # Skip wabdd related files
            if file.name in wabdd_files:
                continue

            # Warn if file is not supported
            if file.suffix not in supported_extensions:
                print(f"WARNING: Skipping {file}, not supported", file=sys.stderr)
                continue

            # Skip ignored files
            if file.suffix in ignored_extensions:
                continue

            queue.put((folder, file))

        overall_download_task_id = download_overall_progress.add_task("Decrypting files...", total=queue.qsize())

        # Start workers
        workers = []
        for _ in range(threads):
            # Add a sentinel to the queue to indicate the end of the queue
            queue.put(_sentinel)

            worker = DecryptionWorker(
                queue,
                output,
                (download_overall_progress, overall_download_task_id),
                key,
            )
            worker.start()
            workers.append(worker)

        # Wait for all workers to finish
        try:
            for worker in workers:
                worker.join()
        except KeyboardInterrupt:
            print("Keyboard Interrupt received! Stopping workers...")
            _stop_event.set()

            # Wait for remaining workers to finish
            for worker in workers:
                worker.join()



# Copyright 2025 Giacomo Ferretti
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import gpsoauth


class AuthException(Exception):
    def __init__(self, message: str, extra: object):
        super().__init__(message)
        self.extra = extra

    def get_extra(self):
        return self.extra


def get_auth_token(master_token: str, android_id: str):
    oauth_data = {
        "accountType": "HOSTED_OR_GOOGLE",
        "has_permission": 1,
        "Token": master_token,
        "service": "oauth2:https://www.googleapis.com/auth/drive.appdata",
        "source": "android",
        "androidId": android_id,
        "app": "com.whatsapp",
        "client_sig": "38a0f7d505fe18fec64fbf343ecaaaf310dbd799",
        "device_country": "us",
        "operatorCountry": "us",
        "lang": "en",
        "sdk_version": 17,
        "google_play_services_version": 240913000,
    }
    auth_response = gpsoauth._perform_auth_request(oauth_data, None)

    # Check if the login was successful
    if "Auth" not in auth_response:
        raise AuthException("Login failed", auth_response)

    return auth_response["Auth"]
